% This function is designed to solve a 1d ode test problem
% so that we can use the concepts learned to write a more in
% depth code later

% The ODE test problem that we are trying to solve is given by:
% a1*(du/dt) - a2*u'' + a3*u + a4*u^2 = f(x)

function [x,uh,t,U,J,resid,resids,l2_resids] = nlwavelet_ode(N,level,dt,T,a1,a2,a3,a4,plotyes,convergence_results)
    
% 
%     
    close all
    
    % The convergence criteria for our newton iteration 
    conv_crit = 1e-6;
    
	% Run the geometry routine to get info about the boundary
	[xl,xr,ul,ur] = geometry(0);
    x_exact = xl:.01:xr;
    
    % Information about the time stepping
    % If a1=0 then do not step through time
    if(a1 == 0)
        nt = 1;
    else
        nt = floor(T./dt);
    end
    
    % The computation time
    %t_comp = 0;

	% Check to make sure the user(you) isnt an idiot
    % And change this to != if in octave, ~= if in MATLAB
	if(mod(xr-xl,(1/(2^level))) ~= 0)
		error('The size of the boundary is not consistent with the Discretization used');
	end

	% Define the grid of points that we are going to use
	x = (xl:(1/(2^level)):xr)';
    
	% The number of scaling functions to use 
	% Also could be thought of as the discretization
	Nx = (xr-xl)./(1/(2^level)) + 1;
    
    % The size of our linear system
    nJ = Nx+2.*(N-1);
    
    % The solution at time = 0
    ukm1 = zeros(nJ,1);
    ukm1(N:nJ-N+1) = u_exact(x,0);
    
%     ukm1 = zeros(Nx,1);
%     ukm1 = u_exact(x,0);
    
    %
	% Determine the Connection coefficients
    %
        
    % for the double derivative
	[~,~,Omega2,~] = connection_coeffs(N,2,0,level);
	%[~,~,Omega2] = Zhang_Connection_coefficients(N,2,0,level);
    
    % for the three term connection coeff
	[Omega3,sucess] = connection_coeffs3(N,0,0,0,level);
	%[~,~,Omega1] = Zhang_Connection_coefficients(N,1,0,level)
    
    % Since the first term is -u'' multiply Omega by -1
	Omega2 = Omega2';
    
    % X modified represents the entire domain including the phantom nodes
    % at each end of the boundary
    xm = ((xl-((N-1)/(2^level))):(1/(2^level)):(xr+((N-1)/(2^level))))';
    
    % Plot the solutions
    if(plotyes == 1)
        % Define a more resolved x domain for the exact solution
        x_exact = xl:.01:xr;
        figure(1)
        hold on
        %plot(x,uh,'g');
        xlabel('x','FontSize',16);
        ylabel('u(x,t)','FontSize',16);
        title(strcat('Wavelet Galerkin solution $\Delta x =  \frac{1}{2^',num2str(level),'}$'),'FontSize',16,'Interpreter','latex');
        %title(strcat('WG vs Actual $\Delta x = \frac{1}{2^',num2str(level),'}\: \: \Delta t = ',num2str(dt),'$'),'FontSize',16,'Interpreter','latex');
        set(gca,'FontSize',16)
    end
    
    % Initial time starts at zero
    t = 0;
    
    % The matrix containing all the values of the solution at all time
    U = zeros(nJ,nt);
    
    % The vector holding the l2residuals
    l2_resids = zeros(20,nt);
    
    % The vector holding the l2residuals
    resids = zeros(nJ,20);

    % Determine the time instances
    time_instances = 0:dt:T;
    nt = length(time_instances)-1;
    quartert = ceil(nt./4);
    halft = floor(nt./2);
    quartert3 = ceil(nt.*.75);
    
    for i=1:nt
        
        % The time at this time step
        t = t + dt;
        
        if(convergence_results == 1)
            fprintf('==============================================================\n');
            fprintf('Time: %d Newton Iteration:\n',i*dt);
        end
                
        % To approximate uk just use the zero vector to start out
        uk = ukm1;
        
        % Set the BC's so that they stick regardless
        uk(N) = ul;
        uk(nJ-N+1) = ur;
        
%         uk(1) = ul;
%         uk(Nx) = ur;
        
        
        
        % Calculate the initial residual
        resid = residual( N, nJ, Omega2, Omega3, uk, ukm1, x, t, dt, a1, a2, a3, a4);

        % Find the norm of the residual for convergence and stuff
        l2resid = norm(resid(N:nJ-N+1));
        %l2resid = norm(resid);
        %l2resid=1;
        
        resids(:,1) = resid;
        l2_resids(1,i) = l2resid;
        
        % Counter to make sure we dont go haywire
        loop_count = 1;

        % This is the non-linear iteration loop where I will use Newton's
        % method to try to converge to an acceptable answer
        while(l2resid > conv_crit && loop_count<20)

            % Finding the Jacobian
            [J] = assemble_jacobian(N, nJ, Omega2, Omega3, dt, a1, a2, a3, a4 );
            
            %  Solve the linear system.
            du = J\(-resid);

            % Use du and uk to find uk+1
            %uk(N:nJ-N+1) = uk(N:nJ-N+1) + du(N:nJ-N+1);
            uk = uk + du;
            
  
            % Set the BC's so that they stick regardless
            uk(N) = ul;
            uk(nJ-N+1) = ur;
%             uk(1) = ul;
%             uk(Nx) = ur;
            
            % Put the solution into the matrix
            U(:,loop_count) = uk;

            % Calculate the residual based on the u and f values
            resid = residual( N, nJ, Omega2, Omega3, uk, ukm1, x, t, dt, a1, a2, a3, a4);
            resids(:,loop_count+1) = resid;
            l2_resids(loop_count+1,i) = l2resid;
            
            % Find the 2 norm of our residual
            %l2resid = norm(resid(N:nJ-N+1));
            l2resid = norm(resid);

            if(convergence_results == 1)
                % Print the residual at each time step
                fprintf('L_2 residual: %e\n',l2resid);
            end

            % Increment the loop count
            loop_count = loop_count + 1;
        end
        
        % Extract the actual domain solution
        uh = uk(N:N+Nx-1);
        
        % Put the uh into the solution matrix
        %U(:,i) = uh;%(2:length(uh)-1);
        
        % Plot the solutions
        if(plotyes == 1)
            
            if((i==1) || (i == nt) || (i == halft) || (i == quartert) || (i == quartert3))
                % Define a more resolved x domain for the exact solution
                
                hold on
                %plot(x,uh,'*-','LineWidth',1.5)
                plot(x,uh,'LineWidth',1.5)
                plot(x_exact,u_exact(x_exact,t),'r','LineWidth',1.5)
                %legend('u_{wavelet}','u_{exact}');
                hold off
            end
        end      
        
        % Set the new time step as the old one
        %ukm1 = uk;
        ukm1(N:nJ-N+1) = uk(N:nJ-N+1);
        
    end
    
    if(plotyes == 1)
        figure(2)
        semilogy(l2_resids(:,1:6),'LineWidth',1.5)
        hold on
        xlabel('Iteration','FontSize',16);
        ylabel('L_2 Residual norm','FontSize',16);
        title('L_2 residual norms with respect to iteration','FontSize',16);
        legend('t=0','t=0.01','t=0.02','t=0.03','t=0.04','t=0.05');
        set(gca,'FontSize',16)
        hold off

        figure(3)
        hold on
        xlabel('x','FontSize',16);
        ylabel('Residual','FontSize',16);
        title('Residuals at the nodes at the final time step','FontSize',16,'Interpreter','latex');
        plot(xm,resids(:,1:5),'LineWidth',1.5)
        legend('1st iter','2nd iter','3rd iter','4th iter','5th iter');
        set(gca,'FontSize',16)
        hold off
    end
    
    
    % Average time it takes to solve the system
    %t_comp = t_comp./nt;
    
    % We are interested in the condition number of the matrix at the final time step
   % q = cond(A);
    %q=0;  
end
    
%% Assemble the Jacobian based on an input set of coneection coefficients
%
%
%
%

function [J] = assemble_jacobian(N, nJ, Omega2, O3, dt, a1, a2, a3, a4 )

    % Initialize the Jacobian
	
    %nA = Nx;
	J = zeros(nJ);

	% Find the middle of our connection coefficients
	nO = (2.*(N-1))-1;
	midO = ceil(nO./2);
    
    % Extract the same shifts from the 3 term cc's to apply to the jacobian
    Omega3 = zeros(1,nO);
    for i=1:nO
        Omega3(i) = O3(i+(nO.*(i-1)));
    end
    
	% Construct the Connection coefficient vector with the added in
	% differential equation constants.  Remember this is the jacobian not
	% the standard matrix in the linear case    

	%
	% Fill in the entries of the matrix
	%
    
	% The Left Boundary
	iJ = [1, midO+1];
	iO = [midO-1,nO];
	for i=2:N-1 
		% Put in the connection coefficients
		J(i,iJ(1):iJ(2)) = a2.*Omega2(iO(1):iO(2)) - 2.*a4.*Omega3(iO(1):iO(2));
        
        % See if this works
        J(i,i) = J(i,i) - ((a1./dt) + a3);
        
		% Shift the Omega and the position in the A
		% matrix
		iJ(2) = iJ(2) + 1;
		iO(1) = iO(1) - 1;

    end
    
%     % The Left Boundary
% 	iJ = [1, midO];
% 	iO = [midO,nO];
% 	for i=2:N 
% 		% Put in the connection coefficients
% 		J(i,iJ(1):iJ(2)) = Omega(iO(1):iO(2));
%         
%         % See if this works
%         J(i,i) = J(i,i) - (2.*a4.*uk(i));
%         
% 		% Shift the Omega and the position in the A
% 		% matrix
% 		iJ(2) = iJ(2) + 1;
% 		iO(1) = iO(1) - 1;
% 
% 	end


	% For the functions that are actually in our interval
	iJ = [2,nO+1];
	for i=N:(nJ-N+1)

        % Put in the connection coeffs
        J(i,iJ(1):iJ(2)) = a2.*Omega2 - 2.*a4.*Omega3;

        % Add in the middle part
        J(i,i) = J(i,i) - ((a1./dt) + a3);

        % Increment the position in the A matrix
        iJ = iJ + 1;

    end
    
    % For the functions that are actually in our interval
%     iJ = [2,nO+1];
% 	for i=N+1:(Nx-midO+1)
% 
%         % Put in the connection coeffs
%         J(i,iJ(1):iJ(2)) = Omega;
% 
%         % Add on the nonlinear contribution
%         J(i,i) = J(i,i) - (2.*a4.*uk(i));
% 
%         % Increment the position in the A matrix
%         iJ = iJ + 1;
% 
% 	end
	
	% The Right boundary
	iJ = [nJ-nO+1,nJ];
	iO = [1, nO];
	for i=(nJ-N+2):(nJ-1)

		% Put in the connection coefficients
		J(i,iJ(1):iJ(2)) = Omega2(iO(1):iO(2)) - 2.*a4.*Omega3(iO(1):iO(2));
        
        % Add in the middle part
        J(i,i) = J(i,i) - ((a1./dt) + a3);
        
		% Shift the Omega and the position in the A
		% matrix
		iJ = [iJ(1)+1,iJ(2)];
		iO = [iO(1), iO(2)-1];

    end
    

%     % The Right boundary
% 	iJ = [Nx-nO+1,Nx];
% 	iO = [1, nO];
% 	for i=(Nx+2-midO):(Nx+1)
% 
% 		% Put in the connection coefficients
% 		J(i,iJ(1):iJ(2)) = Omega(iO(1):iO(2));
% 		
%         % See if this works
%         %J(i,i) = J(i,i) - (2.*a4.*uk(i));
%         
% 		% Shift the Omega and the position in the A
% 		% matrix
% 		iJ = [iJ(1)+1,iJ(2)];
% 		iO = [iO(1), iO(2)-1];
% 
% 	end


	% And now put in the Boundary conditions!
	% This is only for Dirichlet
	J(1,N) = 1;
	J(nJ,nJ-N+1) = 1;
    
%     J(1,2) = 1;
% 	J(Nx+2,Nx-1) = 1;
    
    % If our beta and gamma values are both zero then we just need to
    % change the system a little bit
%     if((beta == 0) && (gamma == 0))
%         if(a1 ~= 0)
%             J = eye(Nx).*(alpha + (a1./dt));
%         else
%              J = eye(Nx).*(alpha);
%         end
%     end



end

%% Determine the residual
%
%
%

function [r] = residual( N, nJ, Omega2, O3, uk, ukm1, x, t, dt, a1, a2, a3, a4)
    
    
    % Extract out all the relevant solution data.  This neglects the zeros
    % that are padded to the begining and end of this vector
    %uk = uk(N:nJ-N+1);
    %ukm1 = ukm1(N:nJ-N+1);
    
    % Now we need to put all the residual values that are multiplied by the
    % delta functions into the vector.  This is everything except for the
    % double derivative term.
    %r(N:nJ-N+1) = ((a1./dt).*ukm1(N:nJ-N+1)) - (((a1./dt) + a3 + (a4.*uk(N:nJ-N+1))).*uk(N:nJ-N+1));
%     [uk, ukm1]
%     rhs(x,t)
%     asdf

    r = ((a1./dt).*ukm1) - (((a1./dt) + a3).*uk);
    %r(2:nJ-1) = r(2:nJ-1) + rhs(x,t);
    r(N:nJ-N+1) = r(N:nJ-N+1) + rhs(x,t);
    %r = r + rhs(x,t);
    %fprintf('part 1: %d\n',r(N));
    
	% Find the middle of our connection coefficients
	nO = (2.*(N-1))-1;
	midO = ceil(nO./2);
          
    % Extract the same shifts from the 3 term cc's to apply to the jacobian
    Omega3 = zeros(1,nO);
    for i=1:nO
        Omega3(i) = O3(i+(nO.*(i-1)));
    end
    
    %
    % Now we need to apply the term that is dependant on the connection
    % coefficients
    
    % First factor in the left boundary
    % The Left Boundary
	ir = [1, midO];
	iO = [midO,nO];
	for i=1:N-1 
		% Put in the connection coefficients
		r(ir(1):ir(2)) = r(ir(1):ir(2)) ... 
                         + uk(ir(1):ir(2)).*a2.*Omega2(iO(1):iO(2))' ...
                         - (uk(ir(1):ir(2)).^2).*a4.*Omega3(iO(1):iO(2))';
        
		% Shift the Omega and the position in the A
		% matrix
		ir(2) = ir(2) + 1;
		iO(1) = iO(1) - 1;
    end
    
%     ir = [1, midO];
% 	iO = [midO,nO];
% 	for i=1:midO-1 
% 		% Put in the connection coefficients
% 		r(ir(1):ir(2)) = r(ir(1):ir(2)) + uk(ir(1):ir(2)).*O(iO(1):iO(2))';
%         
% 		% Shift the Omega and the position in the A
% 		% matrix
% 		ir(2) = ir(2) + 1;
% 		iO(1) = iO(1) - 1;
%     end
    
    % The actual domain
    for i=N:nJ-N+1
        r(i-midO+1:i+midO-1) = r(i-midO+1:i+midO-1) ...
                               + uk(i).*a2.*Omega2' ...
                               - (uk(i).^2).*a4.*Omega3';
    end
    
%     for i=midO:Nx-midO
%         r(i-midO+1:i+midO-1) = r(i-midO+1:i+midO-1) + uk(i).*O';
%     end
%     
    % The Right boundary
	ir = [nJ-nO+1,nJ];
	iO = [1, nO];
	for i=(nJ-N+2):nJ
		% Put in the connection coefficients
		r(ir(1):ir(2)) = r(ir(1):ir(2)) ...
                         + uk(ir(1):ir(2)).*a2.*Omega2(iO(1):iO(2))' ...
                         - (uk(ir(1):ir(2)).^2).*a4.*Omega3(iO(1):iO(2))';
		
		% Shift the Omega and the position in the A
		% matrix
		ir = [ir(1)+1,ir(2)];
		iO = [iO(1), iO(2)-1];

    end

%     ir = [Nx-nO+1,Nx];
% 	iO = [1, nO];
% 	for i=(Nx-midO+1):Nx
% 		% Put in the connection coefficients
% 		r(ir(1):ir(2)) = r(ir(1):ir(2)) + uk(ir(1):ir(2)).*O(iO(1):iO(2))';
% 		
% 		% Shift the Omega and the position in the A
% 		% matrix
% 		ir = [ir(1)+1,ir(2)];
% 		iO = [iO(1), iO(2)-1];
% 
%     end

%fprintf('part 2: %d\n',r(N));
    % Re zero the elements outside of the actual domain
    %r(1) = 0;
    %r(nJ) = 0;
    r(1:N) = 0;
    r(nJ-N+1:nJ) = 0;
    
    %r = [0;r;0];
    
end
    
    


