program nlwavelet_galerkin

	! This program is designed to solve the nonlinear solution to the PDE:
	! a1*(du/dt) - a2*(d^2u/dx^2) + a3*u + a4*(u^2) = f(x,t)
	
	implicit none
	
	! Lets Define the variables we are going to need:
		
	! The max size of the vectors that we could imagine
	integer (kind=4), parameter ::n_alloc = 25000
	! The scaling function genus we plan to use
	integer (kind=4), parameter :: N = 6
	! Effectively the resolution of the grid we will solve on nodes spaced by 2^level
    integer (kind=4), parameter :: level = 6
    ! The temporal discretization
	real (kind=8), parameter :: dt = 0.01d+0
	! The time to run our solution to
	real (kind=8), parameter :: T = 0.05d+0
    ! The Coefficients in front of our PDE
    real (kind=8), parameter :: a1 = 1.0d+0
	real (kind=8), parameter :: a2 = 1.0d+0
	real (kind=8), parameter :: a3 = 1.0d+0
	real (kind=8), parameter :: a4 = 1.0d+0
	
	! The convergence criteria of our newton method
	real (kind=8), parameter :: conv_crit = 1e-6
	! The number of quadrature points to use for the integration
	integer (kind=4), parameter :: show_convergence_rates = 1
	
    ! The filename that the solution will be written to
    character (len=80) :: SOLUTION_csv_file_name = 'Solution.csv'
	integer (kind=4) csv_file_unit
    character (len=n_alloc) record
    
	
	!
	! Program specific variables
	!
	
	! Geometry and Boundary conditions
	real (kind=8) xl
	real (kind=8) xr
	real (kind=8) ul
	real (kind=8) ur
	real (kind=8) x(1:n_alloc)
    real (kind=8) time
	
	! The connection coefficients that will need
	real (kind=8) Omega(1:n_alloc)
	
	! The solutions that we are going to need to store
	real (kind=8) uk(1:n_alloc)
	real (kind=8) ukm1(1:n_alloc)
	
	! The residual of our newton iteration
	real (kind=8) resid(1:n_alloc)
    
    ! The l2 norm of our residual
    real (kind=8) l2resid
	
	! The Jacobian with its diagonals stored in the columns
	real (kind=8) Jacobian(1:n_alloc,1:n_alloc)
	
    ! The number of time steps
    integer (kind=4) nt
	! The number of scaling functions within our domain
	integer (kind=4) Nx
	! The size of the linear system
	integer (kind=4) nJ
	! The number of connection coefficients and the middle of the CC's
	integer (kind=4) nO
	integer (kind=4) midO
	
    ! Our functions return real scalars
    real (kind=8) u_exact
    real (kind=8) L2_norm
        
	! These are just extra variables that we may need
	integer (kind=4) i
	integer (kind=4) j
	integer (kind=4) k
	integer (kind=4) loop_count
	integer (kind=4) sucess !Whether the solve suceeded or not
    integer (kind=4) ipiv(1:n_alloc)
    
	
    ! Open the solution file for writing
    call csv_file_open_write(SOLUTION_csv_file_name, csv_file_unit)
	
	! Define the Geometry of the problem
	call geometry(level, N, Nx, nJ, x, xl, xr, ul, ur)
    
	! Information about the connection coefficients
	nO = (2*(N-1))-1
	midO = (nO/2)+1
	
	! Determine the connection coefficients that we are going to need to 
	! solve this problem
	call CC_2(N, level, 2, nO, Omega)
    
    ! Time is initially at zero
    time = 0.0
    
	! All values of the solution outside of our domain should be zero
    ukm1(1:nJ) = 0.0d+0
	
    ! But inside our domain need to set the initial solution
    do i=1,Nx
        ukm1(N-1+i) = u_exact(x(i),time)
    end do
	
    ! Initialize a record to keep track of the solution
    record = ''
    
    ! First put in the x value points for the domain
    do i = 1,Nx
        call csv_record_append_r8(x(i), record)
    end do
    
    ! Write to the solution file
    call csv_file_record_write(SOLUTION_csv_file_name, csv_file_unit, record)
    
    record = ''
    
    ! Append the solution to the record
    do i = N, nJ-N+1
        call csv_record_append_r8(ukm1(i), record)
    end do
    
    ! Write to the solution file
    call csv_file_record_write(SOLUTION_csv_file_name, csv_file_unit, record)
    
	! Set the temporal discretization
    if(a1 == 0d+0) then
        nt = 1
    else
        nt = T/dt
    end if
	
	! Loop over all the time steps
	do i=1,nt
	
		! Time at this particular time step
        time = time + dt
		
		! Write that we are starting a new time step
		if(show_convergence_rates == 1) then
			write(*,*) '================================================='
			write(*,'(A8,g14.6,A20)') 'Time: ',time,'Newton Iteration:'
		end if
		
		! Set the guess as the previous time step iteration
		uk(1:nJ) = ukm1(1:nJ) 

		! Set the boundary conditions so they stick regardless
		uk(1) = ul
		uk(nJ) = ur
        
		! Compute the residual
		call residual(N, Nx, nJ, nO, midO, Omega, &
                      uk, ukm1, x, time, dt, &
                      a1, a2, a3, a4, &
                      resid)
		
		! Compute the norm of the residual vector
		l2resid = L2_norm(nJ, resid)
		
		! Keep track of the iterations it takes to converge
		loop_count = 1
		
		! This is the newton iteration that attempts to zero out the residual
		do while((l2resid > conv_crit) .AND. (loop_count < 5))
			
			! Assemble the Jacobian
			call assemble_jacobian(N, level, Nx, nJ, nO, midO, Omega, dt, &
                                   uk, a1, a2, a3, a4, Jacobian)
		
            
			! Use DGBSV to solve a general banded system du is the solution
			call dgbsv(nJ, midO, midO, 1, Jacobian, &
                       2*midO+midO+1, ipiv, resid, nJ, sucess)

			! Compute the next iteration solution
			uk(1:nJ) = uk(1:nJ) + resid(1:nJ) 
			
			! Compute the residual
			call residual(N, Nx, nJ, nO, midO, Omega, &
                      uk, ukm1, x, time, dt, &
                      a1, a2, a3, a4, &
                      resid)
			
			! Compute the norm of the residual vector
			l2resid = L2_norm(nJ, resid)
			
			! If we are to write the convergence rates out
			if(show_convergence_rates == 1) then
				write(*,'(A16,g14.6)') 'L_2 residual: ',l2resid
			end if
			
			loop_count = loop_count + 1
		end do
        
        ! Set the solution for the next time step
        record = ''
    
        ! Write the new solution to the file
        do j=N,nJ-N+1
            call csv_record_append_r8(uk(j), record)
        end do
        
    
        call csv_file_record_write(SOLUTION_csv_file_name, csv_file_unit, record)
		
	
	end do
	
    ! Close the solution file
	call csv_file_close_write(SOLUTION_csv_file_name, csv_file_unit)
	
	! Use this for printing things
	!write(*,*) 'The Following is output from the Geometry Routine:'
	!write(*,'(A,i2)') 'Number of nodes:', n
	!write(*,'(A8,A8,A16)') 'node','x','y'
	!do i = 1, n
	!	write ( *, '(2x,i8,2x,g14.6,2x,g14.6)' )  i, x(1,i), x(2,
	!end do
		
		
		
end
		
subroutine assemble_Jacobian(N, level, Nx, nJ, nO, midO, Omega, dt, &
                             uk, a1, a2, a3, a4, Jacobian)

	! This subroutine is intended to construct the Jacobian, J, at a given
	! time step that can be solved using the general banded solver
	implicit none
    
    
	! Inputs
	integer (kind=4) N
	integer (kind=4) level
	integer (kind=4) Nx
	integer (kind=4) nJ
	integer (kind=4) nO
	integer (kind=4) midO
	real (kind=8) Omega(1:nO)
	real (kind=8) dt
	real (kind=8) uk(1:nJ)
	real (kind=8) a1
	real (kind=8) a2
	real (kind=8) a3
	real (kind=8) a4
	
	! Outputs
	real (kind=8) Jacobian(1:3*midO+1,1:nJ)
	
	! Routine specific values
    real (kind=8) temp_middle
	integer (kind=4) i
	integer (kind=4) j
	integer (kind=4) k
	
	! Zero out the jacobian to start out
	!do i=1:3*midO+1
	!	do j=1:nJ
	!		J(i,j) = 0.0
	!	end do
	!end do
	Jacobian(1:3*midO+1,1:nJ) = 0.0d+0
	
	! The most super diagonal just has a one as its firs entry and 
	! is zero everywhere else
	Jacobian(midO+1,N) = 1.0d+0
	Jacobian(3*midO+1,nJ-N+1) = 1.0d+0
	
	! Loop over all the diagonals that we must account for
	do i=1,nO
		
		! if we are on the middle diagonal we must do something special
		if(i == midO) then
			
			! Adjust the middle value so we do not need to keep re-computing
			temp_middle = a2*Omega(midO) - ((a1/dt) + a3)
			
			! Loop over all the values
			do j=2,nJ-1
				
				! Put the value into the matrix
				Jacobian(midO+1+i,j) = temp_middle
				
				! if we are in the interior of the boundary, factor in the 
				! previous iterates guess
				if((j >= N) .and. (j<=(nJ-N+1))) then
					
					Jacobian(midO+1+i,j) = Jacobian(midO+1+i,j) - (2*a4*uk(j))
				end if
			end do
		
		! Otherwise we just put the omega values into the Jacobian
		else
			
			! Just put the values into the entire row since only the
			! ones necessary will be counted.
			Jacobian(midO+1+i,1:nJ) = a2*Omega(i)
			
			! if it is a super diagonal put the zero in the front
			if(i<midO) then
				Jacobian(midO+1+i,midO+1-i) = 0.0d+0
			else if(i>midO) then
				Jacobian(midO+1+i,nJ-(i-midO)) = 0.0d+0
			end if
		end if
	end do
		
	! We now should have J set up in the form used by dgbsv to solve
    
    ! Write out the matrix
    ! Open a file
    open (unit = 8, file = "A.txt")
    
    ! Debugging
    do k = 1,3*midO+1
        do j = 1, nJ
            write(8,'(2x,g14.6)',advance='NO') Jacobian(k,j)
        end do
        write(8,*)
        !write(*,'(2x,g14.6)') init_rhs(i)
    end do
    close (8)
end


subroutine residual(N, Nx, nJ, nO, midO, Omega, &
                    uk, ukm1, x, time, dt, &
                    a1, a2, a3, a4, &
                    resid)
    
    ! This routine is used to compute the residual of our nonlinear system
    ! so that we can do a newton iteration to successively refine our guess
    ! at the actual solution at a given time step
    implicit none
    
    ! Inputs
	integer (kind=4) N
	integer (kind=4) Nx
	integer (kind=4) nJ
	integer (kind=4) nO
	integer (kind=4) midO
	real (kind=8) Omega(1:nO)
	real (kind=8) uk(1:nJ)
	real (kind=8) ukm1(1:nJ)
	real (kind=8) x(1:Nx)
	real (kind=8) time
	real (kind=8) dt
	real (kind=8) a1
	real (kind=8) a2
	real (kind=8) a3
	real (kind=8) a4
	
	! Outputs
	real (kind=8) resid(1:nJ)
    
    ! Routine specific variables
	real (kind=8) rhs
    integer (kind=4) i
    integer (kind=4) j
    integer (kind=4) k

    ! First initialize the residual vector
    resid(1:nJ) = 0.0d+0
    
    ! Now we need to put all the residual values that are multiplied by the
    ! delta functions into the vector.  This is everything except for the
    ! double derivative term.
    do i=N,nJ-N+1
        !write(*,*) rhs(x(i-(N-1)),time)
        resid(i) = rhs(x(i-(N-1)),time) &
                   - (((a1/dt) + a3 + (a4*uk(i)))*uk(i)) &
                   + ((a1/dt)*ukm1(i))
    end do
    
    ! Now we need to apply the term that is dependant on the connection
    ! coefficients
    do i=N,nJ-N+1
        k = 1
        do j=i-midO+1,i+midO-1
            resid(j) = resid(j) + uk(i)*a2*Omega(k);
            k = k + 1
        end do
    end do
    
    ! Convert the residual to negative because that is how a newton iteration is done
    do i=1,nJ
        resid(i) = -resid(i)
    end do
    
    ! Write out the residual vector
    ! Open a file
    open (unit = 9, file = "rhs.txt")
    
    ! Debugging
    do j = 1, nJ
        write(9,'(2x,g14.6)') resid(j)
    end do
    close (9)
    
    
end

	
subroutine geometry(level, N, Nx, nJ, x, xl, xr, ul, ur)

	! This function is intended to declare the grid that we will be working on
	! using information on the level of the discretization and parameters 
	! defined within this subroutine
    implicit none
	
	! The level of discretization
	integer (kind=4) level
	
	! The scaling function genus
	integer (kind=4) N
	! The number of scaling functions in the system
	integer (kind=4) Nx
	! The size of the linear system
	integer (kind=4) nJ
	
	! Geometry and Boundary conditions
	real (kind=8) xl
	real (kind=8) xr
	real (kind=8) ul
	real (kind=8) ur
	real (kind=8) x(1:Nx)	
	
	! The functions we will be using
	real(kind=8) u_exact
	
	! Some routine specific values
	integer (kind=4) i
	real (kind=8) dx
	
	! Define the boundary values
	xl = 0.0d+0
	xr = 1.0d+0
	ul = u_exact(xl,0.0d+0)
	ur = u_exact(xr,0.0d+0)

	! Determine the number of scaling functions to use in the domain
	! This can also be thought of as the number of nodes in the system
	Nx = int((xr-xl)/(1/(2**real(level))) + 1)

	! Find the size of the linear system
	nJ = int(Nx+(2*(N-1)))

	! Determine the grid spacing
	dx = (xr-xl)/(2**level)
	
	! The first value in the x array is xl
	x(1) = xl
	
	! Now create the x domain 
	do i=2,Nx
		x(i) = x(i-1) + dx
	end do

end		
		
function L2_norm(n, vec)
    
    ! This function computes the L2 norm of a vector of size n
    implicit none
    
    ! The inputs
    integer (kind=4) n
    real (kind=8) vec(1:n)
    
    ! The output
    real (kind=8) L2_norm
    
    ! Routine specific variables
    integer (kind=4) i
    real (kind=8) vec_sum
    
    ! Initialize the summation
    vec_sum = 0.0d+0
    
    ! Loop over all the elements
    do i=1,n
        vec_sum = vec_sum + (vec(i)**2)
    end do
	
    ! Take the sqrt of the summation
    L2_norm = sqrt(vec_sum)
    
    return
end    
		
function u_exact(x,time)

	! This is the exact solution of our differential equation,
	! Assuming that it is known
	implicit none
	
	! The inputx value and time
	real (kind=8) x
	real (kind=8) time

	! The output value
	real (kind=8) u_exact
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359d+0

	! Determine the exact solution
	u_exact = time*x*(x-1.0)*(sin(50*x)+(100*time*cos(50*x)))
	
	
	return
end
	
	
function rhs(x,time)

	! This is the right hand side of our newton iteration
	implicit none
	
	! The inputx value and time
	real (kind=8) x
	real (kind=8) time

	! The output value
	real (kind=8) rhs
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359d+0

	! Determine the exact solution
	rhs = x*(x-1)*(sin(50*x)+ 200*time*cos(50*x)) &
          - time*(2*(sin(50*x)+100*time*cos(50*x)) + 100*(x-1)*(cos(50*x)-100*time*sin(50*x)) &
          + 100*x*(cos(100*x)-100*time*sin(50*x)) + (50**2)*x*(x-1)*(-sin(50*x)-100*time*cos(50*x))) &
          + time*x*(x-1)*(sin(50*x)+100*time*cos(50*x)) &
          + ((time*x*(x-1)*(sin(50*x)+100*time*cos(50*x)))**2)
	
	
	return
end	
		
		
		