% This script will evaluate the error of our wavelet galerkin method
% as we increase the level

% The ODE test problem that we are trying to solve is given by:
% eta*(du/dt) - beta*u'' + gamma*u' + alpha*u = f(x)

clear
clc
close all

% The Daubecheis wavelets we would like to test
Nmin = 6;
Nmax = 20;

% how many levels to test
nlevels = 5;

% This is the discretization
h = zeros(nlevels,1);

% The error vector
L2Err = zeros(nlevels,1);
LInfErr = zeros(nlevels,1);


% The Rate of convergence vector
Rate = zeros(nlevels-1,1);

% The value of the constants we would like to use
eta = 1;
beta = 1;
gamma = 0;
alpha = 1;

% The time to run our simulation to
T = 1;

% For all the wavelets
for N = Nmin:2:Nmax
    
    % Print out some of the table info so we can copy/paste
    % it into latex
    fprintf(strcat('D',num2str(N)));
    fprintf('\\begin{center}\n');
    fprintf('\\begin{tabular}{| c | c | c | c |}\n');
    fprintf('\\hline');
    %fprintf('$h$ & Euclidean Error & L2 CR & LInf CR & Condition Number & Average Computation time\\\\\n');
    fprintf('$h$ & Euclidean Error & Euclidean Convergence Rate & Condition Number\\\\\n');
    fprintf('\\hline\n');
    
    % The levels we would like to test
    for level = 2:nlevels;
        
        % The current discretization
        h(level) = 1/(2^level);
        
        % Make sure that the dt is small enough
        dt = h(level).^2;
        %dt = 0.001;
        
        % restrict delta t to be no less than 10^-
        
        % Run the test case we would like to solve
        [x,uh,q,t] = wavelet_ode(N,level,dt,T,eta,beta,gamma,alpha,0);
        
        
        % Get the exact solution
        uex = u_exact(x,t);
        
        % The residual
        r = abs(uex-uh);
        
        % Find the error
        LInfErr(level) = norm(r,Inf)./norm(uex,Inf);
        L2Err(level) = norm(r)./norm(uex);
        
        % Print the row in tabulated format
        if(level > 2)
            L2Rate = (log(L2Err(level-1)./L2Err(level)))./(log(h(level-1))./log(h(level)));
            LInfRate = (log(LInfErr(level-1)./LInfErr(level)))./(log(h(level-1))./log(h(level)));
            %fprintf('%f & %e & %f & %f & %e & %f\\\\\n',h(level),L2Err(level),L2Rate,LInfRate,0,t_comp);
            fprintf('%f & %e & %f & %e\\\\\n',h(level),L2Err(level),L2Rate,q);
        else
            %fprintf('%f & %e & & & %e & %f\\\\\n',h(level),L2Err(level),0,t_comp);
            fprintf('%f & %e & & %e\\\\\n',h(level),L2Err(level),q);
        end
    end
    
    % Finish off the bottom part of the table
    fprintf('\\hline\n');
    fprintf('\\end{tabular}\n');
    fprintf('\\end{center}\n');
end


% figure(1)
% hold on
% plot((1:nlevels),Err,'LineWidth',1.5);
% xlabel('level');
% ylabel('Error');
% title('Error with respect to the level of discretization');
% hold off
% 	
