% This function is designed to solve a 1d ode test problem
% so that we can use the concepts learned to write a more in
% depth code later

% The ODE test problem that we are trying to solve is given by:
% eta*(du/dt) - beta*u'' + gamma*u' + alpha*u = f(x)

function [x,uh,q,t,U,U_ex] = wavelet_ode(N,level,dt,T,eta,beta,gamma,alpha,plotyes)
    
%     N = 6
%     level = 4
%     dt = 1
%     T = 1
%     eta = 0
%     beta = 1
%     gamma = 1
%     alpha = 1
%     plotyes = 1
%     
    close all
    
	% Run the geometry routine to get info about the boundary
	[xl,xr,ul,ur] = geometry(0);
x_exact = xl:.01:xr;
    % Information about the time stepping
    % If eta=0 then do not step through time
    if(eta == 0)
        nt = 1;
    else
        nt = floor(T./dt);
    end
    
    % The computation time
    t_comp = 0;

	% Check to make sure the user(you) isnt an idiot
    % And change this to != if in octave, ~= if in MATLAB
	if(mod(xr-xl,(1/(2^level))) ~= 0)
		error('The size of the boundary is not consistent with the Discretization used');
	end

	% Define the grid of points that we are going to use
	x = (xl:(1/(2^level)):xr)';
    
	% The number of scaling functions to use 
	% Also could be thought of as the discretization
	Nx = (xr-xl)./(1/(2^level)) + 1;

    
    %
	% Determine the Connection coefficients
    %
        
    % for the double derivative
	[temp,temp,Omega2,M2] = connection_coeffs(N,2,0,level);
	%[~,~,Omega2] = Zhang_Connection_coefficients(N,2,0,level);
    
    % for the single derivative
	[temp,temp,Omega1,~] = connection_coeffs(N,1,0,level);
	%[~,~,Omega1] = Zhang_Connection_coefficients(N,1,0,level)

    % Since the first term is -u'' multiply Omega by -1
    Omega = -beta.*Omega2 + gamma.*Omega1;
	Omega = Omega';

	% Initialize the A matrix
	nA = Nx+2.*(N-1);
    %nA = Nx;
	A = zeros(nA);
    
    % X modified represents the entire domain including the phantom nodes
    % at each end of the boundary
    xm = ((xl-((N-2)/(2^level))):(1/(2^level)):(xr+((N-2)/(2^level))))';

	% Find the middle of our connection coefficients
	nO = (2.*(N-1))-1;
	midO = ceil(nO./2);

	% Add in the middle terms
	Omega(midO) = Omega(midO) + alpha;
    
    % If we have a time term neet to add on eta./dt
    if(eta ~= 0)
        Omega(midO) = Omega(midO) + (eta./dt);
    end

	%
	% Fill in the entries of the matrix
	%

    % Formulate the differentiation matrix
    %[A] = make_Diff_matrix(N,nA,Omega);
    
	% The Left Boundary
	iA = [1, midO+1];
	iO = [midO-1,nO];
	for i=2:N-1 
		% Put in the connection coefficients
		A(i,iA(1):iA(2)) = Omega(iO(1):iO(2));
        
		% Shift the Omega and the position in the A
		% matrix
		iA(2) = iA(2) + 1;
		iO(1) = iO(1) - 1;

	end

	% For the functions that are actually in our interval
	iA = [2,nO+1];
	for i=N:(N+Nx-1)

		% Put in the connection coeffs
		A(i,iA(1):iA(2)) = Omega;
		 
		 % Increment the position in the A matrix
		 iA = iA + 1;

	end
	
	% The Right boundary
	iA = [nA-nO+1,nA];
	iO = [1, nO];
	for i=(N+Nx):(nA-1)

		% Put in the connection coefficients
		A(i,iA(1):iA(2)) = Omega(iO(1):iO(2));
		
		% Shift the Omega and the position in the A
		% matrix
		iA = [iA(1)+1,iA(2)];
		iO = [iO(1), iO(2)-1];

	end

	% And now put in the Boundary conditions!
	% This is only for Dirichlet
	A(1,N) = 1;
	A(nA,N+Nx-1) = 1;
    
    % If our beta and gamma values are both zero then we just need to
    % change the system a little bit
    if((beta == 0) && (gamma == 0))
        if(eta ~= 0)
            A = eye(Nx).*(alpha + (eta./dt));
        else
             A = eye(Nx).*(alpha);
        end
    end
        
    % 
    % Now that we have our A matrix setup we can step through time
    % the only thing that changes is the right hand side
    %
    
    % The solution at time = 0
    %uh = zeros(Nx,1);
    uh = u_exact(x,0);
    
    % Plot the solutions
    if(plotyes == 1)
        % Define a more resolved x domain for the exact solution
        x_exact = xl:.01:xr;
        figure(1)
        hold on
        %plot(x,uh,'g');
        xlabel('x','FontSize',16);
        ylabel('u(x,t)','FontSize',16);
        title(strcat('Wavelet Galerkin solution $\Delta x =  \frac{1}{2^',num2str(level),'}$'),'FontSize',16,'Interpreter','latex');
        %title(strcat('WG vs Actual $\Delta x = \frac{1}{2^',num2str(level),'}\: \: \Delta t = ',num2str(dt),'$'),'FontSize',16,'Interpreter','latex');
        set(gca,'FontSize',16)
    end
    
    
    % Initialize the rhs
	b = zeros(nA,1);
    
    
    
    % Initial time starts at zero
    t = 0;
    
    % The matrix containing all the values of the solution at all time
    U = zeros(length(x)-2,nt);
    
    % Determine the time instances
    time_instances = 0:dt:T;
    nt = length(time_instances)-1;
    quartert = ceil(nt./4);
    halft = floor(nt./2);
    quartert3 = ceil(nt.*.75);
    
    for i=1:nt
        
        % The time at this time step
        t = t + dt;
        
        % Perform the wavelet transform to get the rhs in wavelet space
%         [T,cij] = map_to_scaling_fn_space(rhs(x,t),N,level);
%         A = T*A;
        
        % The boundary conditions
        [~,~,ul,ur] = geometry(t);
        b(1) = ul;
        b(nA) = ur;

        % Update the right hand side
        if((beta == 0) && (gamma == 0))
            b = rhs(x,t);
        else
            b(N:N+Nx-1) = rhs(x,t);
            
            % If we have a time term
            if(eta ~= 0)
                b(N:N+Nx-1) = b(N:N+Nx-1) + ((eta./dt).*uh);
            end
            
        end
        
       
        %b = rhs(x);
        
        tic
        
        % Solve the linear system
        uj = A\b;
        
        
        t_comp = t_comp + toc;
    
        % Construct the Approximated Solution
        if((beta == 0) && (gamma == 0))
            uh = uj;
        else
            uh = uj(N:N+Nx-1);
        end
        
        % Put the uh into the solution matrix
        U(:,i) = uh(2:length(uh)-1);
        
        % Transform the solution back to physicalspace
        %uh = T*uh;
        
        % Plot the solutions
        if(plotyes == 1)
            
            if((i==1) || (i == nt) || (i == halft) || (i == quartert) || (i == quartert3))
                % Define a more resolved x domain for the exact solution
                
                hold on
                plot(x,uh,'*-','LineWidth',1.5)
                plot(x,uh,'LineWidth',1.5)
                %plot(x_exact,u_exact(x_exact,t),'r','LineWidth',1.5)
                %legend('u_{wavelet}','u_{exact}');
                hold off
            end
        end
        
        U_ex(:,i) = u_exact(x,t);
        
       
        
    end
    
    % Average time it takes to solve the system
    t_comp = t_comp./nt;
    
    % We are interested in the condition number of the matrix at the final time step
    q = cond(A);
    %q=0;    
    
end
    
    


